@@ -15,16 +15,18 @@ import decimal
 from .compat import PY3, string_base_type
 from .exceptions import ElementPathSyntaxError, ElementPathNameError, MissingContextError
 from .datatypes import UntypedAtomic, DayTimeDuration, YearMonthDuration, \
-    NumericTypeProxy, ArithmeticTypeProxy, XSD_BUILTIN_TYPES
+    NumericTypeProxy, ArithmeticTypeProxy
 from .xpath_context import XPathSchemaContext
 from .tdop_parser import Parser, MultiLabel
 from .namespaces import XML_ID, XML_LANG, XPATH_1_DEFAULT_NAMESPACES, \
     XPATH_FUNCTIONS_NAMESPACE, XSD_NAMESPACE, qname_to_prefixed
+from .schema_proxy import AbstractSchemaProxy
 from .xpath_token import XPathToken
-from .xpath_nodes import AttributeNode, NamespaceNode, is_etree_element, \
-    is_xpath_node, is_element_node, is_document_node, is_attribute_node, \
+from .xpath_nodes import AttributeNode, NamespaceNode, TypedAttribute, TypedElement,\
+    is_etree_element, is_xpath_node, is_element_node, is_document_node, is_attribute_node, \
     is_text_node, is_comment_node, is_processing_instruction_node, node_name
 
+
 XML_NAME_CHARACTER = (u"A-Z_a-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF"
                       u"\u200C\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD")
 XML_NCNAME_PATTERN = u"[{0}][-.0-9\u00B7\u0300-\u036F\u203F-\u2040{0}]*".format(XML_NAME_CHARACTER)
@@ -92,9 +94,9 @@ class XPath1Parser(Parser):
         self.variables = dict(variables if variables is not None else [])
         self.strict = strict
 
-    @classmethod
-    def build_tokenizer(cls, name_pattern=XML_NCNAME_PATTERN):
-        super(XPath1Parser, cls).build_tokenizer(name_pattern)
+    @staticmethod
+    def create_tokenizer(symbol_table, name_pattern=XML_NCNAME_PATTERN):
+        return Parser.create_tokenizer(symbol_table, name_pattern)
 
     @property
     def version(self):
@@ -233,82 +235,65 @@ literal('(string)')
 literal('(float)')
 literal('(decimal)')
 literal('(integer)')
-literal('(name)', bp=10)
 
 
-@method('(name)')
+@method(literal('(name)', bp=10))
 def evaluate(self, context=None):
-    if context is None:
-        return
-    name = self.value
-    if name[0] != '{' and self.parser.default_namespace:
-        name = u'{%s}%s' % (self.parser.default_namespace, name)
-
-    if isinstance(context, XPathSchemaContext):
-        xsd_type = self.match_xsd_type(context.item, name)
-        if xsd_type is not None:
-            if isinstance(context.item, AttributeNode):
-                primitive_type = self.parser.schema.get_primitive_type(xsd_type)
-                return XSD_BUILTIN_TYPES[primitive_type.local_name].value
-            else:
-                return context.item
-
-    elif self.xsd_type is None:
-        if is_attribute_node(context.item, name):
-            return context.item[1]
-        elif is_element_node(context.item, name):
-            return context.item
-    else:
-        try:
-            if is_attribute_node(context.item, name):
-                return self.xsd_type.decode(context.item[1])
-            elif is_element_node(context.item, name):
-                if self.xsd_type.is_simple():
-                    return self.xsd_type.decode(context.item)
-                else:
-                    return context.item
-        except (TypeError, ValueError):
-            self.wrong_context_type("Type %r is not appropriate for the context" % (type(context.item)))
+    return [x for x in self.select(context)]
 
 
 @method('(name)')
 def select(self, context=None):
     if context is None:
         return
+
     name = self.value
-    if name[0] != '{' and self.parser.default_namespace:
-        name = u'{%s}%s' % (self.parser.default_namespace, name)
-
     if isinstance(context, XPathSchemaContext):
-        for item in context.iter_children_or_self():
-            xsd_type = self.match_xsd_type(item, name)
-            if xsd_type is not None:
-                if isinstance(context.item, AttributeNode):
-                    primitive_type = self.parser.schema.get_primitive_type(xsd_type)
-                    yield XSD_BUILTIN_TYPES[primitive_type.local_name].value
-                else:
-                    yield context.item
+        # Bind with the XSD type from a schema
+        if name[0] != '{' and self.parser.default_namespace:
+            name = '{%s}%s' % (self.parser.default_namespace, name)
+
+        for schema_item in context.iter_children_or_self():
+            if self.match_xsd_type(schema_item, name) is not None:
+                yield self.get_typed_node(context, schema_item)
+        return
+
+    if name[0] != '{' and self.parser.default_namespace:
+        tag = '{%s}%s' % (self.parser.default_namespace, name)
+    else:
+        tag = name
+
+    # Checks if the token is bound to an XSD type. If not try a match using
+    # the element path. If this match fails the xsd_type attribute is set
+    # with the schema object to prevent other checks until the schema change.
+    if self.xsd_type is self.parser.schema:
 
-    elif self.xsd_type is None:
         # Untyped selection
         for item in context.iter_children_or_self():
             if is_attribute_node(item, name):
-                yield item[1]
-            elif is_element_node(item, name):
                 yield item
-    else:
-        # Typed selection
+            elif is_element_node(item, tag):
+                yield item
+
+    elif self.xsd_type is None or isinstance(self.xsd_type, AbstractSchemaProxy):
+
+        # Try to match the type using the path
         for item in context.iter_children_or_self():
-            try:
-                if is_attribute_node(item, name):
-                    yield self.xsd_type.decode(item[1])
-                elif is_element_node(item, name):
-                    if self.xsd_type.is_simple():
-                        yield self.xsd_type.decode(item)
-                    else:
-                        yield item
-            except (TypeError, ValueError):
-                self.wrong_sequence_type("Type %r does not match sequence type of %r" % (self.xsd_type, item))
+            if is_attribute_node(item, name) or is_element_node(item, tag):
+                path = context.get_path(item)
+
+                xsd_component = self.parser.schema.find(path, self.parser.namespaces)
+                if xsd_component is not None:
+                    self.xsd_type = xsd_component.type
+                else:
+                    self.xsd_type = self.parser.schema
+
+                yield self.get_typed_node(context, item)
+    else:
+        # XSD typed selection
+        for item in context.iter_children_or_self():
+            if is_attribute_node(item, name) or is_element_node(item, tag):
+                yield self.get_typed_node(context, item)
 
 
 ###
@@ -385,13 +370,37 @@ def select(self, context=None):
         else:
             value = '{%s}%s' % (namespace, self[1].value)
 
-    if context is not None:
+    if context is None:
+        return
+    elif isinstance(context, XPathSchemaContext):
+        for schema_item in context.iter_children_or_self():
+            if self.match_xsd_type(schema_item, value) is not None:
+                yield self.get_typed_node(context, schema_item)
+
+    elif self.xsd_type is self.parser.schema:
         for item in context.iter_children_or_self():
             if is_attribute_node(item, value):
-                yield item[1]
+                yield item
             elif is_element_node(item, value):
                 yield item
 
+    elif self.xsd_type is None or isinstance(self.xsd_type, AbstractSchemaProxy):
+        for item in context.iter_children_or_self():
+            if is_attribute_node(item, value) or is_element_node(item, value):
+                path = context.get_path(item)
+                xsd_component = self.parser.schema.find(path, self.parser.namespaces)
+                if xsd_component is not None:
+                    self.xsd_type = xsd_component.type
+                else:
+                    self.xsd_type = self.parser.schema
+                yield self.get_typed_node(context, item)
+
+    else:
+        # XSD typed selection
+        for item in context.iter_children_or_self():
+            if is_attribute_node(item, value) or is_element_node(item, value):
+                yield self.get_typed_node(context, item)
+
 
 ###
 # Namespace URI as in ElementPath
@@ -500,26 +509,28 @@ def select(self, context=None):
     if context is None:
         self.missing_context()
     else:
-        try:
-            parent = context.parent_map[context.item]
-        except KeyError:
-            pass
-        else:
-            if is_element_node(parent):
-                context.item = parent
-                yield parent
+        parent = context.get_parent(context.item)
+        if is_element_node(parent):
+            context.item = parent
+            yield parent
 
 
 ###
 # Logical Operators
 @method(infix('or', bp=20))
 def evaluate(self, context=None):
-    return bool(self[0].evaluate(context) or self[1].evaluate(context))
+    if context is None:
+        return self.boolean_value(self[0].evaluate()) or self.boolean_value(self[1].evaluate())
+    return self.boolean_value(self[0].evaluate(context.copy())) or \
+        self.boolean_value(self[1].evaluate(context.copy()))
 
 
 @method(infix('and', bp=25))
 def evaluate(self, context=None):
-    return bool(self[0].evaluate(context) and self[1].evaluate(context))
+    if context is None:
+        return self.boolean_value(self[0].evaluate()) and self.boolean_value(self[1].evaluate())
+    return self.boolean_value(self[0].evaluate(context.copy())) and \
+        self.boolean_value(self[1].evaluate(context.copy()))
 
 
 @method(infix('=', bp=30))
@@ -637,35 +648,49 @@ def evaluate(self, context=None):
 
 ###
 # Union expressions
-@method(infix('|', bp=50))
+@method('|', bp=50)
+def led(self, left):
+    self.cut_and_sort = True
+    if left.symbol in {'|', 'union'}:
+        left.cut_and_sort = False
+    self[:] = left, self.parser.expression(rbp=50)
+    return self
+
+
+@method('|')
 def select(self, context=None):
-    if context is not None:
-        results = {item for k in range(2) for item in self[k].select(context.copy())}
-        for item in context.iter():
-            if item in results:
+    if context is None:
+        return
+    elif not self.cut_and_sort:
+        for k in range(2):
+            for item in self[k].select(context.copy()):
                 yield item
+    else:
+        results = {item for k in range(2) for item in self[k].select(context.copy())}
+        for item in context.iter_results(results):
+            yield item
 
 
 ###
 # Path expressions
-@method('//', bp=80)
-@method('/', bp=80)
+@method('//', bp=75)
+@method('/', bp=75)
 def nud(self):
     next_token = self.parser.next_token
     if next_token.symbol == '(end)' and self.symbol == '/':
         return self
     elif not self.parser.next_is_path_step_token():
         next_token.wrong_syntax()
-    self[:] = self.parser.expression(80),
+    self[:] = self.parser.expression(75),
     return self
 
 
-@method('//', bp=80)
-@method('/', bp=80)
+@method('//')
+@method('/')
 def led(self, left):
     if not self.parser.next_is_path_step_token():
         self.parser.next_token.wrong_syntax()
-    self[:] = left, self.parser.expression(80)
+    self[:] = left, self.parser.expression(75)
     return self
 
 
@@ -684,39 +709,27 @@ def select(self, context=None):
         for result in self[0].select(context):
             yield result
     else:
-        items = set()
-        left_results = list(self[0].select(context))
+        items = []
+        left_results = [x for x in self[0].select(context)]
         context.size = len(left_results)
         for context.position, context.item in enumerate(left_results):
-            if not is_element_node(context.item):
-                self.wrong_type("left operand must returns element nodes: {}".format(context.item))
+            if not is_xpath_node(context.item):
+                self.wrong_type("left operand must returns XPath nodes: {}".format(context.item))
             for result in self[1].select(context):
-                if is_etree_element(result) or isinstance(result, tuple):
-                    if result not in items:
+                if not is_etree_element(result) and not isinstance(result, tuple):
+                    yield result
+                elif result in items:
+                    pass
+                elif isinstance(result, (TypedAttribute, TypedElement)):
+                    if result[0] not in items:
+                        items.append(result)
                         yield result
-                        items.add(result)
+                elif isinstance(result, AttributeNode):
+                    items.append(result)
+                    yield result
                 else:
+                    items.append(result)
                     yield result
-
-
-@method('/')
-def evaluate(self, context=None):
-    """
-    General evaluation method for path operators, that may returns the a single value or None.
-    """
-    if context is not None:
-        selector = iter(self.select(context))
-        try:
-            value = next(selector)
-        except StopIteration:
-            return
-        else:
-            try:
-                next(selector)
-            except StopIteration:
-                return self.data_value(value)
-            else:
-                self.wrong_context_type("atomized operand is a sequence of length greater than one")
 
 
 @method('//')
@@ -724,6 +737,7 @@ def select(self, context=None):
     if context is None:
         return
     elif len(self) == 1:
+        context.item = None
         for _ in context.iter_descendants(axis='descendant-or-self'):
             for result in self[0].select(context):
                 yield result
@@ -738,7 +752,7 @@ def select(self, context=None):
 
 ###
 # Predicate filters
-@method('[', bp=75)
+@method('[', bp=80)
 def led(self, left):
     self.parser.next_token.unexpected(']')
     self[:] = left, self.parser.expression()
@@ -749,16 +763,13 @@ def led(self, left):
 @method('[')
 def select(self, context=None):
     if context is not None:
-        left_results = list(self[0].select(context))
-        context.size = len(left_results)
-        for context.position, context.item in enumerate(left_results):
-            predicate = list(self[1].select(context.copy()))
-            if len(predicate) == 1 and not isinstance(predicate[0], bool) and \
-                    isinstance(predicate[0], (int, float)):
-                if context.position == predicate[0] - 1:
-                    yield context.item
+        for position, item in enumerate(self[0].select(context), start=1):
+            predicate = [x for x in self[1].select(context.copy())]
+            if len(predicate) == 1 and isinstance(predicate[0], NumericTypeProxy):
+                if position == predicate[0]:
+                    yield item
             elif self.boolean_value(predicate):
-                yield context.item
+                yield item
 
 
 ###
@@ -820,11 +831,22 @@ def select(self, context=None):
 @method(axis('following-sibling'))
 def select(self, context=None):
     if context is not None:
-        if is_element_node(context.item):
+        if isinstance(context.item, TypedElement):
+            item = context.item[0]
+        elif is_etree_element(context.item):
             item = context.item
-            for elem in context.iter_parent(axis=self.symbol):
+        else:
+            return
+
+        for parent in context.iter_parent(axis=self.symbol):
+            if isinstance(context, XPathSchemaContext):
+                for _ in context.iter_children_or_self(parent, child_axis=True):
+                    for result in self[0].select(context):
+                        yield result
+
+            else:
                 follows = False
-                for child in context.iter_children_or_self(elem, child_axis=True):
+                for child in context.iter_children_or_self(parent, child_axis=True):
                     if follows:
                         for result in self[0].select(context):
                             yield result
@@ -911,7 +933,7 @@ def select(self, context=None):
 def select(self, context=None):
     if context is not None:
         item = context.item
-        for elem in reversed(list(context.iter_ancestors(axis=self.symbol))):
+        for elem in reversed([x for x in context.iter_ancestors(axis=self.symbol)]):
             context.item = elem
             yield elem
         yield item
@@ -934,14 +956,9 @@ def select(self, context=None):
 @method(axis('preceding'))
 def select(self, context=None):
     if context is not None and is_element_node(context.item):
-        elem = context.item
-        ancestors = set(context.iter_ancestors(axis=self.symbol))
-        for e in context.root.iter():
-            if e is elem:
-                break
-            if e not in ancestors:
-                context.item = e
-                yield e
+        for _ in context.iter_preceding():
+            for result in self[0].select(context):
+                yield result
 
 
 ###
@@ -972,9 +989,7 @@ def evaluate(self, context=None):
 def select(self, context=None):
     if context is not None:
         for item in context.iter_children_or_self():
-            if item is None:
-                yield context.root
-            elif is_text_node(item):
+            if is_text_node(item):
                 yield item
 
 
@@ -992,13 +1007,7 @@ def evaluate(self, context=None):
 
 @method(function('count', nargs=1))
 def evaluate(self, context=None):
-    results = self[0].evaluate(context)
-    if isinstance(results, list):
-        return len(results)
-    elif results is not None:
-        return 1
-    else:
-        return 0
+    return len([x for x in self[0].select(context)])
 
 
 @method(function('id', nargs=1))
@@ -1274,4 +1283,4 @@ def evaluate(self, context=None):
 
 
 register('(end)')
-XPath1Parser.build_tokenizer()
+XPath1Parser.build()
