@@ -19,6 +19,7 @@ import time
 import re
 import locale
 import unicodedata
+import xml.etree.ElementTree as ElementTree
 
 from .compat import PY3, string_base_type, unicode_chr, urlparse, urljoin, urllib_quote, unicode_type
 from .datatypes import QNAME_PATTERN, DateTime10, Date10, Time, Timezone, Duration, DayTimeDuration
@@ -182,11 +183,21 @@ def select(self, context=None):
         elem = self.get_argument(context)
         if not is_element_node(elem):
             raise self.error('FORG0006', 'argument %r is not a node' % elem)
-        for e in elem.iter():
-            tag_ns = get_namespace(e.tag)
-            for pfx, uri in self.parser.namespaces.items():
-                if uri == tag_ns:
-                    yield pfx
+
+        if isinstance(context, XPathSchemaContext):
+            # For schema context returns prefixes of static namespaces
+            for prefix in self.parser.namespaces:
+                yield prefix
+        elif hasattr(elem, 'nsmap'):
+            # For lxml returns Element's prefixes
+            for prefix in elem.nsmap:
+                yield prefix or ''
+        else:
+            # For ElementTree returns module registered prefixes
+            prefixes = {x for x in self.parser.namespaces}
+            prefixes.update(x for x in ElementTree._namespace_map.values())
+            for prefix in prefixes:
+                yield prefix
 
 
 @method(function('resolve-QName', nargs=2))
@@ -313,7 +324,7 @@ def evaluate(self, context=None):
 # Aggregate functions
 @method(function('avg', nargs=1))
 def evaluate(self, context=None):
-    result = list(self[0].select(context))
+    result = [x for x in self[0].select(context)]
     if not result:
         return result
     elif isinstance(result[0], Duration):
@@ -425,7 +436,7 @@ def select(self, context=None):
 
 @method(function('reverse', nargs=1))
 def select(self, context=None):
-    for result in reversed(list(self[0].select(context))):
+    for result in reversed([x for x in self[0].select(context)]):
         yield result
 
 
@@ -440,7 +451,7 @@ def select(self, context=None):
 
 @method(function('unordered', nargs=1))
 def select(self, context=None):
-    for result in sorted(list(self[0].select(context)), key=lambda x: self.string_value(x)):
+    for result in sorted([x for x in self[0].select(context)], key=lambda x: self.string_value(x)):
         yield result
 
 
@@ -646,7 +657,7 @@ def evaluate(self, context=None):
 
 @method(function('string-join', nargs=2))
 def evaluate(self, context=None):
-    items = [self.string_value(s) if is_element_node(s) else s
+    items = [self.string_value(s) if is_element_node(s) or is_attribute_node(s) else s
              for s in self[0].select(context)]
     try:
         return self.get_argument(context, 1, cls=string_base_type).join(items)
