@@ -10,9 +10,10 @@
 #
 import datetime
 
+from .compat import lru_cache
 from .exceptions import ElementPathTypeError
-from .xpath_nodes import AttributeNode, is_etree_element, is_element_node, \
-    is_document_node, is_attribute_node
+from .xpath_nodes import AttributeNode, TypedAttribute, TypedElement, is_etree_element, \
+    is_element_node, is_document_node, is_attribute_node
 
 
 class XPathContext(object):
@@ -41,13 +42,12 @@ class XPathContext(object):
             raise ElementPathTypeError(
                 "invalid argument root={!r}, an Element is required.".format(root)
             )
-        self._root = root
+
+        self.root = root
         if item is not None:
             self.item = item
-        elif is_element_node(root):
-            self.item = root
         else:
-            self.item = None
+            self.item = root if hasattr(root, 'tag') else None
 
         self.position = position
         self.size = size
@@ -55,16 +55,17 @@ class XPathContext(object):
         self.variables = {} if variables is None else dict(variables)
         self.current_dt = current_dt or datetime.datetime.now()
         self.timezone = timezone
+        self._elem = item if is_element_node(item) else root
         self._parent_map = None
 
     def __repr__(self):
         return '%s(root=%r, item=%r, position=%r, size=%r, axis=%r)' % (
-            self.__class__.__name__, self._root, self.item, self.position, self.size, self.axis
+            self.__class__.__name__, self.root, self.item, self.position, self.size, self.axis
         )
 
     def copy(self, clear_axis=True):
         obj = type(self)(
-            root=self._root,
+            root=self.root,
             item=self.item,
             position=self.position,
             size=self.size,
@@ -73,19 +74,53 @@ class XPathContext(object):
             current_dt=self.current_dt,
             timezone=self.timezone,
         )
+        obj._elem = self._elem
         obj._parent_map = self._parent_map
         return obj
 
-    @property
-    def root(self):
-        return self._root
-
     @property
     def parent_map(self):
         if self._parent_map is None:
-            self._parent_map = {child: elem for elem in self._root.iter() for child in elem}
+            self._parent_map = {child: elem for elem in self.root.iter() for child in elem}
         return self._parent_map
 
+    @lru_cache(maxsize=1024)
+    def get_parent(self, elem):
+        """
+        Returns the parent element or `None` for root element and for elements
+        that are not included in the tree. Uses a LRU cache to minimize parent
+        map rebuilding for trees processed with an incremental parser.
+        """
+        if isinstance(elem, TypedElement):
+            elem = elem[0]
+        if elem is self.root:
+            return
+
+        try:
+            return self._parent_map[elem]
+        except (KeyError, TypeError):
+            self._parent_map = {child: elem for elem in self.root.iter() for child in elem}
+            try:
+                return self._parent_map[elem]
+            except KeyError:
+                return
+
+    def get_path(self, item):
+        """Cached path resolver for elements and attributes. Returns absolute paths."""
+        path = []
+        if isinstance(item, (AttributeNode, TypedAttribute)):
+            path.append('@%s' % item[0])
+            item = self._elem
+        if isinstance(item, TypedElement):
+            item = item[0]
+
+        while True:
+            parent = self.get_parent(item)
+            path.append(item.tag)
+            if parent is None:
+                return '/{}'.format('/'.join(reversed(path)))
+            item = parent
+
     def is_principal_node_kind(self):
         if self.axis == 'attribute':
             return is_attribute_node(self.item)
@@ -102,32 +137,37 @@ class XPathContext(object):
     def iter_attributes(self):
         if not is_element_node(self.item):
             return
+        self._elem = self.item
 
         status = self.item, self.size, self.position, self.axis
         self.axis = 'attribute'
 
-        for item in self.item.attrib.items():
-            self.item = AttributeNode(*item)
+        if isinstance(self.item, TypedElement):
+            self.item = self.item.elem
+
+        for self.item in map(lambda x: AttributeNode(*x), self.item.attrib.items()):
             yield self.item
 
         self.item, self.size, self.position, self.axis = status
 
     def iter_children_or_self(self, item=None, child_axis=False):
-        status = self.item, self.size, self.position, self.axis
         if not child_axis and self.axis is not None:
             yield self.item
-            self.item, self.size, self.position, self.axis = status
             return
 
+        status = self.item, self.size, self.position, self.axis
         self.axis = 'child'
+
         if item is not None:
-            self.item = item
+            self.item = item[0] if isinstance(item, TypedElement) else item
+        elif isinstance(self.item, TypedElement):
+            self.item = self.item[0]
 
         if self.item is None:
             self.size, self.position = 1, 0
-            self.item = self._root.getroot() if is_document_node(self._root) else self._root
+            self.item = self.root.getroot() if is_document_node(self.root) else self.root
             yield self.item
-        elif is_element_node(self.item):
+        elif is_etree_element(self.item):
             elem = self.item
             if elem.text is not None:
                 self.item = elem.text
@@ -138,29 +178,61 @@ class XPathContext(object):
 
         self.item, self.size, self.position, self.axis = status
 
-    def iter_parent(self, axis=None):
+    def iter_preceding(self):
+        item = e = self.item[0] if isinstance(self.item, TypedElement) else self.item
+        if not is_etree_element(item):
+            return
+
         status = self.item, self.size, self.position, self.axis
-        self.axis = axis
-        try:
-            self.item = self.parent_map[self.item]
-        except KeyError:
-            pass
-        else:
-            yield self.item
+        self.axis = 'preceding'
+
+        ancestors = []
+        while True:
+            parent = self.get_parent(e)
+            if parent is None:
+                break
+            else:
+                ancestors.append(parent)
+                e = parent
+
+        for e in self.root.iter():
+            if e is item:
+                break
+            elif e not in ancestors:
+                self.item = e
+                yield e
+
         self.item, self.size, self.position, self.axis = status
 
+    def iter_parent(self, axis=None):
+        if isinstance(self.item, TypedElement):
+            parent = self.get_parent(self.item[0])
+        else:
+            parent = self.get_parent(self.item)
+
+        if parent is not None:
+            status = self.item, self.size, self.position, self.axis
+            self.axis = axis
+
+            self.item = parent
+            yield self.item
+
+            self.item, self.size, self.position, self.axis = status
+
     def iter_descendants(self, item=None, axis=None):
         status = self.item, self.size, self.position, self.axis
         self.axis = axis
 
         if item is not None:
-            self.item = item
+            self.item = item[0] if isinstance(item, TypedElement) else item
+        elif isinstance(self.item, TypedElement):
+            self.item = self.item[0]
 
         if self.item is None:
             self.size, self.position = 1, 0
-            yield self._root
-            self.item = self._root.getroot() if is_document_node(self._root) else self._root
-        elif not is_etree_element(self.item):
+            yield self.root
+            self.item = self.root.getroot() if is_document_node(self.root) else self.root
+        elif not is_element_node(self.item):
             return
 
         for descendant in self._iter_descendants():
@@ -169,7 +241,7 @@ class XPathContext(object):
         self.item, self.size, self.position, self.axis = status
 
     def _iter_descendants(self):
-        elem = self.item
+        elem = self._elem = self.item
         yield elem
         if elem.text is not None:
             self.item = elem.text
@@ -185,14 +257,13 @@ class XPathContext(object):
         self.axis = axis
 
         if item is not None:
-            self.item = item
-        if not is_etree_element(self.item):
-            return
+            self.item = item[0] if isinstance(item, TypedElement) else item
+        elif isinstance(self.item, TypedElement):
+            self.item = self.item[0]
 
         while True:
-            try:
-                parent = self.parent_map[self.item]
-            except KeyError:
+            parent = self.get_parent(self.item)
+            if parent is None:
                 break
             else:
                 self.item = parent
@@ -206,8 +277,10 @@ class XPathContext(object):
 
         if self.item is None:
             self.size, self.position = 1, 0
-            yield self._root
-            self.item = self._root.getroot() if is_document_node(self._root) else self._root
+            yield self.root
+            self.item = self.root.getroot() if is_document_node(self.root) else self.root
+        elif isinstance(self.item, TypedElement):
+            self.item = self.item[0]
         elif not is_etree_element(self.item):
             return
 
@@ -216,16 +289,36 @@ class XPathContext(object):
 
         self.item, self.size, self.position, self.axis = status
 
+    def iter_results(self, results):
+        """Iterates results in document order."""
+        status = self.item, self.size, self.position
+        self.item = self.root
+
+        for item in self._iter_context():
+            if item in results:
+                yield item
+            elif isinstance(item, AttributeNode):
+                # Match XSD decoded attributes
+                for attr in filter(lambda x: isinstance(x, TypedAttribute), results):
+                    if attr[0] == item:
+                        yield attr
+            elif is_etree_element(item):
+                # Match XSD decoded elements
+                for elem in filter(lambda x: isinstance(x, TypedElement), results):
+                    if elem[0] is item:
+                        yield elem
+
+        self.item, self.size, self.position = status
+
     def _iter_context(self):
-        elem = self.item
+        elem = self._elem = self.item
         yield elem
         if elem.text is not None:
             self.item = elem.text
             yield self.item
 
-        for item in elem.attrib.items():
-            self.item = item
-            yield item
+        for self.item in map(lambda x: AttributeNode(*x), elem.attrib.items()):
+            yield self.item
 
         if len(elem):
             self.size = len(elem)
