@@ -26,6 +26,12 @@ AttributeNode = namedtuple('Attribute', 'name value')
 NamespaceNode = namedtuple('Namespace', 'prefix uri')
 """A namedtuple-based type to represent XPath namespaces."""
 
+TypedAttribute = namedtuple('TypedAttribute', 'attr value')
+"""A wrapper for processing typed-value attributes."""
+
+TypedElement = namedtuple('TypedElement', 'elem value')
+"""A wrapper for processing typed-value elements."""
+
 
 ###
 # Utility functions for ElementTree's Element instances
@@ -34,6 +40,9 @@ def is_etree_element(obj):
 
 
 def elem_iter_strings(elem):
+    if isinstance(elem, TypedElement):
+        elem = elem.elem
+
     for e in elem.iter():
         if e.text is not None:
             yield e.text
@@ -60,9 +69,12 @@ def is_element_node(obj, tag=None):
     :param tag: a fully qualified name, a local name or a wildcard. The accepted wildcard formats \
     are '*', '*:*', '*:local-name' and '{namespace}*'.
     """
-    if not is_etree_element(obj) or callable(obj.tag):
+    if isinstance(obj, TypedElement):
+        obj = obj.elem
+    elif not is_etree_element(obj) or callable(obj.tag):
         return False
-    elif tag is None:
+
+    if not tag:
         return True
     elif not obj.tag:
         return obj.tag == tag
@@ -93,10 +105,13 @@ def is_attribute_node(obj, name=None):
     are '*', '*:*', '*:local-name' and '{namespace}*'.
     """
     if name is None or name == '*' or name == '*:*':
-        return isinstance(obj, AttributeNode)
-    elif not isinstance(obj, AttributeNode):
+        return isinstance(obj, (AttributeNode, TypedAttribute))
+    elif not isinstance(obj, (AttributeNode, TypedAttribute)):
         return False
-    elif name[0] == '*':
+    elif isinstance(obj, TypedAttribute):
+        obj = obj.attr
+
+    if name[0] == '*':
         try:
             _, _name = name.split(':')
         except (ValueError, IndexError):
@@ -116,11 +131,11 @@ def is_schema_node(obj):
 
 
 def is_comment_node(obj):
-    return is_etree_element(obj) and callable(obj.tag) and obj.tag.__name__ == 'Comment'
+    return hasattr(obj, 'tag') and callable(obj.tag) and obj.tag.__name__ == 'Comment'
 
 
 def is_processing_instruction_node(obj):
-    return is_etree_element(obj) and callable(obj.tag) and obj.tag.__name__ == 'ProcessingInstruction'
+    return hasattr(obj, 'tag') and callable(obj.tag) and obj.tag.__name__ == 'ProcessingInstruction'
 
 
 def is_document_node(obj):
@@ -140,8 +155,8 @@ else:
 
 
 def is_xpath_node(obj):
-    return isinstance(obj, tuple) or is_etree_element(obj) or \
-        is_document_node(obj) or is_text_node(obj) or is_schema_node(obj)
+    return isinstance(obj, tuple) or is_etree_element(obj) or is_schema_node(obj) or \
+        is_document_node(obj) or is_text_node(obj)
 
 
 ###
