@@ -19,19 +19,22 @@ In XPath there are 7 kinds of nodes:
 Element-like objects are used for representing elements and comments, ElementTree-like objects
 for documents. Generic tuples are used for representing attributes and named-tuples for namespaces.
 """
+from __future__ import unicode_literals
 import locale
 import contextlib
 from decimal import Decimal
 
-from .compat import string_base_type
+from .compat import string_base_type, unicode_type
 from .exceptions import xpath_error
 from .namespaces import XQT_ERRORS_NAMESPACE
-from .xpath_nodes import AttributeNode, is_etree_element, is_attribute_node, \
-    elem_iter_strings, is_text_node, is_namespace_node, is_comment_node, \
-    is_processing_instruction_node, is_element_node, is_document_node, \
-    is_xpath_node, is_schema_node
+from .xpath_nodes import AttributeNode, TypedAttribute, TypedElement, \
+    is_etree_element, is_attribute_node, elem_iter_strings, is_text_node, \
+    is_namespace_node, is_comment_node, is_processing_instruction_node, \
+    is_element_node, is_document_node, is_xpath_node, is_schema_node
 from .datatypes import UntypedAtomic, Timezone, DayTimeDuration, XSD_BUILTIN_TYPES
+from .schema_proxy import AbstractSchemaProxy
 from .tdop_parser import Token
+from .xpath_context import XPathSchemaContext
 
 
 def ordinal(n):
@@ -51,9 +54,8 @@ def ordinal(n):
 
 class XPathToken(Token):
     """Base class for XPath tokens."""
-
-    comment = None   # for XPath 2.0+ comments
-    xsd_type = None  # fox XPath 2.0+ schema types labeling
+    comment = None    # for XPath 2.0+ comments
+    xsd_type = None   # fox XPath 2.0+ schema types labeling
 
     def evaluate(self, context=None):
         """
@@ -61,7 +63,7 @@ class XPathToken(Token):
 
         :param context: The XPath dynamic context.
         """
-        return list(self.select(context))
+        return [x for x in self.select(context)]
 
     def select(self, context=None):
         """
@@ -95,23 +97,25 @@ class XPathToken(Token):
     def source(self):
         symbol, label = self.symbol, self.label
         if label == 'axis':
-            return u'%s::%s' % (self.symbol, self[0].source)
+            return '%s::%s' % (self.symbol, self[0].source)
         elif label in ('function', 'constructor'):
-            return u'%s(%s)' % (self.symbol, ', '.join(item.source for item in self))
+            return '%s(%s)' % (self.symbol, ', '.join(item.source for item in self))
         elif symbol == ':':
-            return u'%s:%s' % (self[0].source, self[1].source)
+            return '%s:%s' % (self[0].source, self[1].source)
         elif symbol == '(':
-            return '()' if not self else u'(%s)' % self[0].source
+            return '()' if not self else '(%s)' % self[0].source
+        elif symbol == '[':
+            return '%s[%s]' % (self[0].source, self[1].source)
         elif symbol == ',':
-            return u'%s, %s' % (self[0].source, self[1].source)
+            return '%s, %s' % (self[0].source, self[1].source)
         elif symbol == '$':
-            return u'$%s' % self[0].source
+            return '$%s' % self[0].source
         elif symbol == '{':
-            return u'{%s}%s' % (self[0].value, self[1].value)
+            return '{%s}%s' % (self[0].value, self[1].value)
         elif symbol == 'instance':
-            return u'%s instance of %s' % (self[0].source, ''.join(t.source for t in self[1:]))
+            return '%s instance of %s' % (self[0].source, ''.join(t.source for t in self[1:]))
         elif symbol == 'treat':
-            return u'%s treat as %s' % (self[0].source, ''.join(t.source for t in self[1:]))
+            return '%s treat as %s' % (self[0].source, ''.join(t.source for t in self[1:]))
         return super(XPathToken, self).source
 
     @property
@@ -205,7 +209,7 @@ class XPathToken(Token):
         for item in self.select(context):
             value = self.data_value(item)
             if value is None:
-                raise self.error('FOTY0012', "argument node does not have a typed value: {}".format(item))
+                raise self.error('FOTY0012', "argument node {!r} does not have a typed value".format(item))
             else:
                 yield value
 
@@ -225,7 +229,17 @@ class XPathToken(Token):
             try:
                 next(selector)
             except StopIteration:
-                return str(value) if isinstance(value, UntypedAtomic) else value
+                if isinstance(value, UntypedAtomic):
+                    value = str(value)
+                if isinstance(context, XPathSchemaContext):
+                    return value
+                if self.xsd_type is not None and isinstance(value, string_base_type):
+                    try:
+                        value = self.xsd_type.decode(value)
+                    except (TypeError, ValueError):
+                        msg = "Type {!r} is not appropriate for the context"
+                        self.wrong_context_type(msg.format(type(value)))
+                return value
             else:
                 self.wrong_context_type("atomized operand is a sequence of length greater than one")
 
@@ -240,10 +254,11 @@ class XPathToken(Token):
         :returns: a list of data couples.
         """
         if context is None:
-            operand1, operand2 = list(self[0].select()), list(self[1].select())
+            operand1 = [x for x in self[0].select()]
+            operand2 = [x for x in self[1].select()]
         else:
-            operand1 = list(self[0].select(context.copy()))
-            operand2 = list(self[1].select(context.copy()))
+            operand1 = [x for x in self[0].select(context.copy())]
+            operand2 = [x for x in self[1].select(context.copy())]
 
         if self.parser.compatibility_mode:
             # Boolean comparison if one of the results is a single boolean value (1.)
@@ -267,6 +282,22 @@ class XPathToken(Token):
         return [(self.data_value(value1), self.data_value(value2))
                 for value1 in operand1 for value2 in operand2]
 
+    def select_results(self, context):
+        """
+        Generates formatted XPath results.
+
+        :param context: the XPath dynamic context.
+        """
+        for result in self.select(context):
+            if isinstance(result, TypedElement):
+                yield result[0]
+            elif isinstance(result, AttributeNode):
+                yield result[1]
+            elif isinstance(result, TypedAttribute):
+                yield result[0][1] if hasattr(result[0][1], 'type') else result[1]
+            else:
+                yield result
+
     def get_results(self, context):
         """
         Returns formatted XPath results.
@@ -275,7 +306,7 @@ class XPathToken(Token):
         :return: a list or a simple datatype when the result is a single simple type \
         generated by a literal or function token.
         """
-        results = list(self.select(context))
+        results = [x for x in self.select_results(context)]
         if len(results) == 1:
             res = results[0]
             if isinstance(res, (bool, int, float, Decimal)):
@@ -355,12 +386,9 @@ class XPathToken(Token):
         schema type an exception is raised.
 
         :param schema_item: an XPath item related with a schema instance.
-        :param name: a not empty string.
+        :param name: a QName in extended format for matching the item.
         :returns: the matched XSD type or `None` if there isn't a match.
         """
-        if name[0] != '{' and self.parser.default_namespace:
-            name = '{%s}%s' % (self.parser.default_namespace, name)
-
         if isinstance(schema_item, AttributeNode):
             if not schema_item[1].is_matching(name):
                 return
@@ -396,6 +424,39 @@ class XPathToken(Token):
             self.wrong_context_type("Multiple XSD type matching during static analysis")
         return xsd_type
 
+    def get_typed_node(self, context, item):
+        """
+        Returns a typed node if the token is bound to an XSD type.
+
+        :param context: the XPath dynamic context.
+        :param item: an untyped XPath attribute ot element.
+        """
+        if isinstance(self.xsd_type, (type(None), AbstractSchemaProxy)):
+            return item
+
+        if isinstance(context, XPathSchemaContext):
+            primitive_type = self.parser.schema.get_primitive_type(self.xsd_type)
+            try:
+                value = XSD_BUILTIN_TYPES[primitive_type.local_name or 'anyType'].value
+            except KeyError:
+                value = XSD_BUILTIN_TYPES['anyType'].value
+
+            if isinstance(item, AttributeNode):
+                return TypedAttribute(item, value)
+            else:
+                return TypedElement(item, value)
+        else:
+            try:
+                if isinstance(item, AttributeNode):
+                    return TypedAttribute(item, self.xsd_type.decode(item[1]))
+                elif self.xsd_type.is_simple() or self.xsd_type.has_simple_content():
+                    return TypedElement(item, self.xsd_type.decode(item.text))
+                else:
+                    return item
+            except (TypeError, ValueError):
+                msg = "Type {!r} does not match sequence type of {!r}"
+                self.wrong_sequence_type(msg.format(self.xsd_type, item))
+
     @contextlib.contextmanager
     def use_locale(self, collation):
         """A context manager for setting a specific collation for a code block."""
@@ -412,20 +473,22 @@ class XPathToken(Token):
             locale.setlocale(locale.LC_ALL, default_locale)
 
     ###
-    # XPath data conversion base functions
+    # XPath data accessors base functions
     def data_value(self, obj):
         """
         The typed value, as computed by fn:data() on each item. Returns an instance of
         UntypedAtomic.
         """
+        if is_attribute_node(obj) or isinstance(obj, TypedElement):
+            obj = obj[1]
+
         if obj is None:
             return
         elif not is_xpath_node(obj):
             return obj
         elif hasattr(obj, 'type'):
-            return self.schema_node_value(obj)
-        else:
-            return UntypedAtomic(self.string_value(obj))
+            return self.schema_node_value(obj)  # Schema context
+        return UntypedAtomic(self.string_value(obj))
 
     def boolean_value(self, obj):
         """
@@ -455,13 +518,13 @@ class XPathToken(Token):
         if obj is None:
             return ''
         elif is_element_node(obj):
-            return u''.join(elem_iter_strings(obj))
+            return ''.join(elem_iter_strings(obj))
         elif is_attribute_node(obj):
-            return obj[1]
+            return unicode_type(obj[1])
         elif is_text_node(obj):
             return obj
         elif is_document_node(obj):
-            return u''.join(e.text for e in obj.getroot().iter() if e.text is not None)
+            return ''.join(e.text for e in obj.getroot().iter() if e.text is not None)
         elif is_namespace_node(obj):
             return obj[1]
         elif is_comment_node(obj):
